module CUDA_Driver

using CUDA_Driver_jll

using Libdl

export libcuda, system_version

# keep track of the system driver's version, for reporting purposes
# XXX: ideally we'd just use NVML.driver_version(),
#      but that reports the forward-compatible one.
const _system_version = Ref{VersionNumber}()

"""
    system_version()

Returns the version of the system driver (whether or not it's loaded).
This can be used for reporting, but serves no other purpose; you really want to check the
actual version of the driver loaded at run-time by calling `cuDriverGetVersion`.
"""
system_version() = isassigned(_system_version) ? _system_version[] : nothing

function driver_version(path)
    handle = Libdl.dlopen(path; throw_error=true)
    handle = handle::Ptr{Nothing} # doesn't const-prop throw_error
    try
        function_handle = Libdl.dlsym(handle, "cuDriverGetVersion")
        version_ref = Ref{Cint}()
        status = ccall(function_handle, Cint, (Ptr{Cint},), version_ref)
        if status != 0
            return nothing
        end
        major, ver = divrem(version_ref[], 1000)
        minor, patch = divrem(ver, 10)
        return VersionNumber(major, minor, patch)
    finally
        Libdl.dlclose(handle)
    end
end

function init_driver(path)
    library_handle = Libdl.dlopen(path; throw_error=true)
    library_handle = library_handle::Ptr{Nothing} # doesn't const-prop throw_error
    function_handle = Libdl.dlsym(library_handle, "cuInit")
    status = ccall(function_handle, Cint, (UInt32,), 0)
    Libdl.dlclose(library_handle)
    return status
end

function find_system_driver()
    driver = if Sys.iswindows()
        Libdl.find_library("nvcuda")
    else
        Libdl.find_library(["libcuda.so.1", "libcuda.so"])
    end

    if driver == ""
        return nothing
    else
        return driver
    end
end

function choose_driver()
    # find the local system library
    system_driver = find_system_driver()
    if system_driver === nothing
        return nothing
    end

    # query the system driver version
    # XXX: apparently cuDriverGetVersion can be used before cuInit,
    #      despite the docs stating "any function [...] will return
    #      CUDA_ERROR_NOT_INITIALIZED"; is this a recent change?
    system_version = driver_version(system_driver)
    if system_version === nothing
        return nothing
    end
    _system_version[] = system_version
    @debug "System CUDA driver found at $system_driver, detected as version $system_version"

    # check the user preference
    if !parse(Bool, get(ENV, "JULIA_CUDA_USE_COMPAT", "true"))
        @debug "User disallows using forward-compatible driver."
        return system_driver
    end

    # check the version
    if system_version >= CUDA_Driver_jll.version
        @debug "System CUDA driver is recent enough; not using forward-compatible driver"
        return system_driver
    end

    # check if we managed to unload the system driver.
    # if we didn't, we can't consider a forward compatible library because that would
    # risk having multiple copies of libcuda.so loaded (also see NVIDIA bug #3418723)
    system_driver_loaded = Libdl.dlopen(system_driver, Libdl.RTLD_NOLOAD; throw_error=false) !== nothing
    if system_driver_loaded
        @debug "Could not unload the system CUDA library; this prevents use of the forward-compatible driver"
        return system_driver
    end

    # check if this process is hooked by CUDA's injection libraries, which prevents
    # unloading libcuda after dlopening. this is problematic, because we might want to
    # after loading a forwards-compatible libcuda and realizing we can't use it. without
    # being able to unload the library, we'd run into issues (see NVIDIA bug #3418723)
    hooked = haskey(ENV, "CUDA_INJECTION64_PATH")
    if hooked
        @debug "Running under CUDA injection tools; this prevents use of the forward-compatible driver"
        return system_driver
    end

    # check if we even have an artifact
    if !isdefined(CUDA_Driver_jll, :libcuda)
        @debug "No forward-compatible CUDA library available for your platform."
        return system_driver
    end
    compat_driver = CUDA_Driver_jll.libcuda
    @debug "Forward-compatible CUDA driver found at $compat_driver; reported as version $(CUDA_Driver_jll.version)"

    # finally, load the compatibility driver to see if it supports this platform
    # TODO: do we need to dlopen the JIT compiler library for it to be discoverable?
    #       doesn't that clash with a system one if compat cuInit fails? or should
    #       we load it _after_ the compat driver initialization succeeds?
    #compat_compiler = CUDA_Driver_jll.libnvidia_ptxjitcompiler
    #Libdl.dlopen(compat_compiler)

    init_status = init_driver(compat_driver)
    if init_status != 0
        @debug "Could not use forward compatibility package (error $init_status)"

        # see comment above about unloading the system driver
        compat_driver_loaded = Libdl.dlopen(compat_driver, Libdl.RTLD_NOLOAD; throw_error=false) !== nothing
        if compat_driver_loaded
            error("""Could not unload the forward compatible CUDA driver library.
                     This is probably caused by running Julia under a tool that hooks CUDA API calls.
                     In that case, prevent Julia from loading multiple drivers by setting JULIA_CUDA_USE_COMPAT=false in your environment.""")
        end

        return system_driver
    end

    @debug "Successfully loaded forwards-compatible CUDA driver"
    return compat_driver
end

# libcuda cannot be a JLL-style constant string, because we often precompile CUDA-dependent
# packages in an environment where CUDA isn't available (e.g. as part of `docker build`).
# if we afterwards load the package, we want to be able to detect CUDA without recompiling.
const _libcuda = Ref{Union{Nothing,String}}()

"""
    libcuda()

Returns the name/path of the CUDA driver, or `nothing` if no driver is available.
"""
libcuda() = _libcuda[]

function __init__()
    @assert !isassigned(_libcuda)
    _libcuda[] = choose_driver()
    if _libcuda[] !== nothing
        if haskey(ENV, "_") && basename(ENV["_"]) == "rr"
            @error("Running under rr, which is incompatible with CUDA")
            return
        end

        status = @ccall libcuda().cuInit(0::Cint)::Cint
        if status != 0
            error("Could not initialize CUDA driver")
        end
    end
end

end
